// *******************************************************
//
// NAME:    Iceblox
// VERSION:	1.0
// DATE:    2002-06-28
// BY:      Karl Hornell
// 
// RESOURCE: morc res.txt
// COMPILE: pip-gcc -c -O2 -Wall -o iceblox.o iceblox.c
// LINK: pip-gcc iceblox.o res.o -mstack=1024 -mdata=13000 -o iceblox.mpn
//
// *******************************************************

#include <vmgp.h>
#include <vmgputil.h>
#include <vstream.h>
#include "res.h"

__asm(".data\nFONTA::\n.incbin \"myfont.fnt\"");
extern char FONTA[]; 
VMGPFONT		FontA;

__asm(".data\nTILES::\n.incbin \"bloxtilet.raw\"");
extern char TILES[]; 

__asm(".data\nSPRITEDATA::\n.incbin \"bloxsprite.raw\"");
extern char SPRITEDATA[];

typedef struct // Mobile object
{
	int16_t x,y,pos,dir,look,temp;
	int8_t type;
} mobject;

// Constants
	#define		NUMMOBS			10
	#define		SCREEN_HEIGHT	80
	#define		SCREEN_WIDTH	101
	#define 	SPRITELOOKS		33

// Sounds
uint8_t brring[]={
	0,3, 50,0, 80,
	0,4, 50,0, 75,
	0,5, 50,0, 70,
	0,6, 50,0, 65,
	0,7, 50,0, 60,
	0,8, 50,0, 50,
	0,9, 50,0, 35,
	0,10, 50,0, 15
};

uint8_t death[]={
	0,1, 50,0, 80,
	240,0, 50,0, 75,
	220,0, 50,0, 70,
	195,0, 50,0, 65,
	165,0, 50,0, 60,
	130,0, 50,0, 50,
	90,0, 50,0, 35,
	45,0, 50,0, 15,
	25,0, 200,0, 10
};

int highscores[10];

char *versNum="VERSION 1.0",*credits="BY KARL H@RNELL",*toStart="FIRE - START   UP - INFO",
	*gameOver="GAME OVER",*warning="GETTING MORE DIFFICULT",*highscore="HIGHSCORES";
char *infoText[]={
	"YOUR JOB IS TO COLLECT",
	"FIVE GOLD COINS FROZEN",
	"IN ICE. PUSH OR CRUSH",
	"ICE BLOCKS BY WALKING",
	"AGAINST THEM.",
	" "," "," ",
	"WATCH OUT FOR THE EVIL",
	"FLAMES. YOU CAN SNUFF",
	"THEM WITH ICE."};

	MAP_HEADER bghdr;
	SPRITE spritehdr;
	uint8_t *tilemap,*spr[SPRITELOOKS],*logo;
	char tmp[10];
	mobject	*objs;

	uint8_t i,j,k;
	int8_t maxFlames,maxReds,flames,reds,gameMode=1,lastSkill;

void readHighscores()
{
	int res=vResOpenMode(NULL,HIGHSCORE,STREAM_READ);
	if (res>-1)
	{
		for (k=0;k<10;k++)
			vResRead(res,&highscores[k],sizeof(int));
		vResClose(res);
	}
}

void storeHighscores()
{
	int res=vResOpenMode(NULL,HIGHSCORE,STREAM_WRITE);
	if (res>-1)
	{
		for (k=0;k<10;k++)
			vResWrite(res,&highscores[k],sizeof(int));
		vResClose(res);
	}
}

void initResources(void) 
{
	FontA.width		= 4; 
	FontA.height	= 6;
	FontA.bpp		= 1;
	FontA.palindex	= 0;
	FontA.chartbl	= FONTA + 16; 
	FontA.fontdata	= FONTA + 16 + 256;	
	vSetActiveFont(&FontA);
	
	int32_t handle,size;
	handle=vResOpen(NULL,LOGSPR); // Logo
	logo=vNewPtr(LOGSPR_UNPACKED_SIZE);
	size=vDecompress(NULL,logo,handle,1024);
	vStreamClose(handle);
}

void prepareLevel(int8_t map[],int16_t blockNum[],uint8_t *tilemap)
{
	int16_t i,j,k,l,m,pile[100];
	int16_t offsets[]={0,-12,12,-1,1};
	int8_t blockTiles[]={0,1,8,15};
	m=0;
	while (m<5)
	{
		// Randomly place blocks
		for (i=0;i<12;i++)
			for (j=0;j<12;j++)
			{
				k=i*12+j;
				if (i>0 && i<11 && j>0 && j<11)
					map[k]=0;
				else
					map[k]=15;
			}
			map[13]=15;
			map[143]=0;
			for (j=1;j<4;j++)
				for (i=0;i<blockNum[j];i++)
				{
					k=0;
					while (map[k]!=0)
					{
						k=12+(vGetRandom()%120);
						if (map[k]!=0)
							k=0;
						else
						{
							map[k]=blockTiles[j];
							k=143;
						}
					}
				}
			
			// Check whether all coins are reachable -- otherwise the level
			// cannot be solved and needs to be rebuilt
			map[13]=16;
			pile[0]=13;
			i=0;j=1;m=0;
			while (i<j)
			{
				k=pile[i++];
				for (l=1;l<5;l++)
					if (map[k+offsets[l]]<15)
					{
						pile[j++]=k+offsets[l];
						if (map[k+offsets[l]]==8)
							m++;
						map[k+offsets[l]]+=16;
					}
			}
	}
	for (i=0;i<144;i++)
		map[i]&=15;

	for(i=0;i<24;i++)
		for(j=0;j<24;j++)
			tilemap[j+i*24] = 4*map[12*(i/2)+(j/2)]+2*(i&1)+(j&1)+1;
}

void initiateObject(mobject	*objs, int16_t ox, int16_t oy,int16_t op,int16_t od,int8_t ot,int16_t ol)
{
	int16_t i;
	i=0;
	while (objs[i].type>0)
		i++;
	objs[i].x=ox;
	objs[i].y=oy;
	objs[i].pos=op;
	objs[i].dir=od;
	objs[i].type=ot;
	objs[i].look=ol;
}

void alterTile(int16_t x, int16_t y, int16_t t)
{
	vMapSetTile(2*x,2*y,1+4*t);
	vMapSetTile(2*x+1,2*y,2+4*t);
	vMapSetTile(2*x,2*y+1,3+4*t);
	vMapSetTile(2*x+1,2*y+1,4+4*t);
}

void mainLoop(void)
{
	uint8_t sprCount;

	int16_t counter=0,scX=16,scY=16,coins=0,lives=0,l,showW=0,showH=0;
	int8_t map[12*12];
	int8_t cyclic[]={0,1,2,1};
	int16_t blockNum[]={0,27,5,6};
	int score=0,level=0;
	int16_t offsets[]={0,-12,12,-1,1};
	int16_t dx[]={0,0,0,-1,1},dy[]={0,-1,1,0,0};
	uint32_t lastKey=0,motionKeys[]={0,KEY_UP,KEY_DOWN,KEY_LEFT,KEY_RIGHT};
	int16_t introSprites[5]={22,19,3,20,28};

	// Set up the smaller Sprites
	spritehdr.width		= 16;					// Sprite width
	spritehdr.height	= 16;					// Sprite height
	spritehdr.format	= VCAPS_RGB332;			// Graphics format RGB332
	spritehdr.centerx	= 0;					// Center point x = 0
	spritehdr.centery	= 0;					// Center point y = 0


	for (i=0;i<SPRITELOOKS;i++)
	{
		spr[i] = vNewPtr(256+sizeof(SPRITE));
		memcpy(spr[i],&spritehdr,sizeof(SPRITE));
		memcpy(spr[i]+sizeof(SPRITE),SPRITEDATA+256*i,256);
	}
	
	vSetRandom(vGetTickCount()); // init the randomizer
	
	objs=vNewPtr(NUMMOBS*sizeof(mobject)); // Allocate memory for the mobile objects
	
	tilemap	= vNewPtr(24*24); // allocate memory for the tilemap
		  
	// Set up and init the Background engine
	bghdr.x					= 0;					// X position on the map
	bghdr.y					= 0;					// Y position on the map
	bghdr.flag              = 0;					// no flags
	bghdr.xpan              = 0;
	bghdr.ypan              = 0;	
	bghdr.width				= 24;					// Width in tiles
	bghdr.height			= 24;					// Height in tiles 
	bghdr.mapoffset			= tilemap;				// Tile Data
	bghdr.tiledata			= TILES;				// Tile Gfx
	bghdr.format			= VCAPS_RGB332;			// Graphics format RGB332
	 
	vMapInit(&bghdr);
	vSpriteInit(NUMMOBS);
	
	vSetForeColor(vRGB(255,255,255)); // White forecolor
	vSetBackColor(vRGB(0,0,0)); // Black background
	vSetTransferMode(MODE_TRANS); // Set transfer mode to transparent

	while(gameMode>0)
	{
		lastKey=vGetButtonData();
		
		counter=(counter+1)&1023;
		
		switch(gameMode)
		{
			case 1: // Prepare intro
				vSetClipWindow(0,0,SCREEN_WIDTH,SCREEN_HEIGHT);
				vClearScreen(vRGB(0,0,0));
				for (l=0;l<576;l++)
					tilemap[l]=1;
				vSpriteClear();
				vSpriteSet(0,(SPRITE*)logo,(SCREEN_WIDTH-96)/2,1);
				vUpdateSpriteMap();
				vPrint(MODE_BLOCK,(SCREEN_WIDTH-44)/2,30,versNum);
				vPrint(MODE_BLOCK,(SCREEN_WIDTH-60)/2,39,credits);
				vPrint(MODE_BLOCK,(SCREEN_WIDTH-96)/2,SCREEN_HEIGHT-7,toStart);
				vFlipScreen(1);
				vSetClipWindow(0,50,SCREEN_WIDTH,66);
				counter=0;
				gameMode=2;
				break;
			case 2: // During intro, waiting for the player to press a key
				introSprites[0]=21+((counter&6)>>1);
				introSprites[4]=28+cyclic[(counter&6)>>1];
				if ((counter&127)<64)
				{
					introSprites[2]=9+cyclic[(counter&6)>>1];
					for (l=0;l<5;l++)
						vSpriteSet(l+1,(SPRITE*)spr[introSprites[l]],2*(counter&63)+20*l-57,50);
				}
				else
				{
					introSprites[2]=6+cyclic[(counter&6)>>1];
					for (l=0;l<5;l++)
						vSpriteSet(l+1,(SPRITE*)spr[introSprites[l]],70-2*(counter&63)+20*l,50);
				}
				vUpdateSpriteMap();
				vFlipScreen(1);
				if (lastKey==KEY_FIRE)
				{
					// Prepare new game
					vClearScreen(vRGB(0,0,0));
					score=0;
					level=1;
					lives=2;
					lastSkill=1;
					gameMode=3;
				}
				else if (lastKey==KEY_UP)
				{
					vSpriteClear();
					vSetClipWindow(0,0,SCREEN_WIDTH,SCREEN_HEIGHT);
					vClearScreen(vRGB(0,0,0));
					for (i=0;i<11;i++)
						vPrint(MODE_BLOCK,2,2+7*i,infoText[i]);
					vSpriteSet(0,(SPRITE*)spr[20],2,38);
					vSpriteSet(1,(SPRITE*)spr[11],30,38);
					vSpriteSet(2,(SPRITE*)spr[19],65,38);
					vSpriteSet(3,(SPRITE*)spr[22],82,38);
					vUpdateSprite();
					vDrawLine(52,43,62,43);
					vDrawLine(50,47,62,47);
					counter=0;
					gameMode=9;
				}
				break;
			case 3: // Setting up new level
				prepareLevel(map,blockNum,tilemap);
				vMapSetXY(13,13);
				vSpriteClear();
				if (level==1)
				{
					maxFlames=1;
					blockNum[1]=27;blockNum[3]=6;
				}
				else if (level<4)
				{
					maxFlames=2;
					blockNum[1]=25;blockNum[3]=7;
				}
				else if (level<6)
				{
					maxFlames=3;
					blockNum[1]=23;blockNum[3]=8;
				}
				else
				{
					maxFlames=4;
					blockNum[1]=21;blockNum[3]=9;
				}
				if (level<8) // Determine number of red flames
					maxReds=0;
				else if (level<14)
					maxReds=(level-6)/2;
				else
					maxReds=4;
				flames=0;reds=0;
				coins=0;
				for (i=0;i<NUMMOBS;i++)
					objs[i].type=0;
				initiateObject(objs,16,16,13,0,1,3);
				showW=SCREEN_WIDTH;
				showH=SCREEN_HEIGHT;
				while (showW>4 && showH>4)
				{
					showW-=4;
					showH-=4;
				}
				if (lastSkill==maxFlames+maxReds)
					gameMode=4;
				else // Getting more difficult, show enemies
				{
					vSpriteClear();
					vPrint(MODE_BLOCK,(SCREEN_WIDTH-88)/2,35,warning);
					vSpriteSet(0,(SPRITE*)spr[3],(SCREEN_WIDTH-16)/2,10);
					for (i=0;i<maxFlames;i++)
						vSpriteSet(i+1,(SPRITE*)spr[i<maxReds?30:22],
						(SCREEN_WIDTH-maxFlames*18)/2+18*i,47);
					vUpdateSprite();
					counter=0;
					gameMode=10;
				}
				lastSkill=maxFlames+maxReds;
				break;
			case 4: // Entering new level, zooming in
				vSetClipWindow((SCREEN_WIDTH-showW)/2,(SCREEN_HEIGHT-showH)/2,
					(SCREEN_WIDTH-showW)/2+showW,(SCREEN_HEIGHT-showH)/2+showH);
				vUpdateSpriteMap();
				vFlipScreen(1);
				showW+=4;
				showH+=4;
				if (showW>SCREEN_WIDTH)
					gameMode=5;
				break;
			case 5: // Main game loop
				vSpriteClear();
				sprCount=0;
				if ((counter&31)==0 && flames<maxFlames) // Launch new flame
				{
					if (objs[0].x<80) // Appear on right-hand side
					{
						j=23;
						k=3;
					}
					else
					{
						j=12;
						k=4;
					}
					j+=12*(vGetRandom()&7);
					while (j<122 && map[j+offsets[k]]>0)
						j+=12;
					if (j<122) // Found an entrance
					{
						initiateObject(objs,16*(j%12),16*(j/12),j,k,reds<maxReds?7:6,21);
						if (reds<maxReds)
							reds++;
						flames++;
					}
				}
				
				for (i=0;i<NUMMOBS;i++) // Process all mobile objects
				{
					switch(objs[i].type)
					{
						case 1: // Penguin
							if ((objs[i].x&15)+(objs[i].y&15)==0)
							{
								objs[i].dir=0;
								for (j=1;j<5;j++)
									if (lastKey==motionKeys[j])
									{
										k=objs[i].pos+offsets[j];
										if (map[k]==0)
										{
											objs[i].dir=j;
											objs[i].pos+=offsets[j];
										}
										else if (map[k]<15) // Bumped into block
										{
											if (map[k+offsets[j]]==0) // Push?
											{
												initiateObject(objs,objs[i].x+16*dx[j],
													objs[i].y+16*dy[j],
													objs[i].pos+offsets[j],
													j,
													map[k]==1?2:3,
													map[k]==1?19:20);
												map[k]=0;
												alterTile(k%12,k/12,0);
											}
											else // Crush?
											{
												if (map[k]==8) // Frozen coin
													vPlayResource(brring,40,SOUND_TYPE_BEEP);

												initiateObject(objs,-50,0,
													objs[i].pos+offsets[j],
													0,
													map[k]==1?4:5,
													map[k]==1?1:8);
												map[k]=15;
											}
										}
									}
							}
							if (objs[i].dir>0)
							{
								objs[i].x+=2*dx[objs[i].dir];
								objs[i].y+=2*dy[objs[i].dir];
								objs[i].look=3*objs[i].dir-3+cyclic[(counter&7)>>1];
							}
							
							// Scroll screen to follow the penguin
							scX=objs[i].x+8-SCREEN_WIDTH/2;
							scY=objs[i].y+8-SCREEN_HEIGHT/2;
							if (scX<13)
								scX=13;
							else if (scX>179-SCREEN_WIDTH)
								scX=179-SCREEN_WIDTH;
							if (scY<13)
								scY=13;
							else if (scY>179-SCREEN_HEIGHT)
								scY=179-SCREEN_HEIGHT;
								
							if (coins==5) // All coins taken?
							{
								flames=50;
								for (j=0;j<NUMMOBS;j++)
									if ((objs[j].type&14)==6)
										objs[j].type=0;
								objs[i].type=10;
								counter=0;
							}		
							break;
						case 2: // Moving block
						case 3: // Moving frozen coin
							objs[i].x+=4*dx[objs[i].dir];
							objs[i].y+=4*dy[objs[i].dir];
							for (j=1;j<NUMMOBS;j++) // Check flame collision
								if ((objs[j].type&14)==6 && objs[j].x<objs[i].x+13 &&
									objs[i].x<objs[j].x+13 && objs[j].y<objs[i].y+13 &&
									objs[i].y<objs[j].y+13)
								{
									if (objs[j].type==6) // Ordinary flame
									{
										objs[j].look=25;
										score+=50;
									}
									else
									{
										objs[j].look=31;
										score+=100;
									}
									objs[j].type=8;
									objs[j].temp=0; // Cycles before disappearing
									vBeep(350,90);
								}
							if ((objs[i].x&15)+(objs[i].y&15)==0)
							{
								objs[i].pos+=offsets[objs[i].dir];
								k=objs[i].pos;
								if (map[k+offsets[objs[i].dir]]>0) // Stop?
								{
									j=objs[i].type==2?1:8;
									map[k]=j;
									alterTile(k%12,k/12,j);
									objs[i].type=0;
								}
							}
							break;
						case 4: // Crushed plain ice
						case 5: // Crushed frozen coin
							if ((counter&1)==0)
							{
								objs[i].look++;
								if (objs[i].look!=8 && objs[i].look!=15)
								{
									alterTile(objs[i].pos%12,objs[i].pos/12,objs[i].look);
								}
								else
								{
									map[objs[i].pos]=0;
									alterTile(objs[i].pos%12,objs[i].pos/12,0);
									if (objs[i].type==5) // Coin?
									{
										coins++;
										score+=100;
									}
									else // Plain ice
										score+=5;
									objs[i].type=0;
								}
							}
							break;
						case 6: // Flame
						case 7: // Fast red flame
							if (objs[i].type==6)
							{
								objs[i].look=21+(counter&6)/2;
								objs[i].x+=dx[objs[i].dir];
								objs[i].y+=dy[objs[i].dir];
							}
							else
							{
								objs[i].look=28+cyclic[(counter&6)/2];
								objs[i].x+=2*dx[objs[i].dir];
								objs[i].y+=2*dy[objs[i].dir];
							}
							if ((objs[i].x&15)+(objs[i].y&15)==0)
							{
								objs[i].pos+=offsets[objs[i].dir];

								if (objs[i].dir==0 || (counter&14)==0)
									objs[i].dir=1+(vGetRandom()&3);
								// Possibly home in on the penguin
								if (objs[i].x<objs[0].x+10 && objs[i].x>objs[0].x-10)
								{
									if (objs[i].y<objs[0].y)
										objs[i].dir=2;
									else
										objs[i].dir=1;
								}
								else if (objs[i].y<objs[0].y+10 && objs[i].y>objs[0].y-10)
								{
									if (objs[0].x>objs[i].x)
										objs[i].dir=4;
									else
										objs[i].dir=3;
								}
								// Stop if there is something in the way
								if (map[objs[i].pos+offsets[objs[i].dir]]>0)
									objs[i].dir=0;
							}
							
							// Caught penguin?
							if (objs[i].x<objs[0].x+9 && objs[i].x>objs[0].x-9 &&
								objs[i].y<objs[0].y+9 && objs[i].y>objs[0].y-9)
							{
								objs[0].type=9;
								objs[0].look=13;
								counter=0;
								flames=50;
								for (j=0;j<NUMMOBS;j++)
									if ((objs[j].type&14)==6)
										objs[j].type=0;
								vPlayResource(death,45,SOUND_TYPE_BEEP);
							}
							break;
						case 8: // Flashing score symbol
							objs[i].look+=(counter&3)/2;
							if (objs[i].look==27)
								objs[i].look=25;
							else if (objs[i].look==33)
								objs[i].look=31;
							objs[i].temp++;
							if (objs[i].temp>15)
							{
								objs[i].type=0;
								flames--;
								if (objs[i].look>27)
									reds--;
							}
							break;
						case 9: // Dying penguin
							if ((counter&1)==0)
							{
								if (objs[i].look<18)
									objs[i].look++;
							}
							if (counter>55)
							{
								lives--;
								if (lives>=0)
								{
									flames=0;reds=0;
									objs[0].type=1;
									objs[0].look=3;
								}
								else
									gameMode=6; // Game over
							}
							break;
						case 10: // Happy penguin
							objs[i].look=3+9*((counter&2)>>1);
							if (counter>75)
							{
								showW=SCREEN_WIDTH;
								showH=SCREEN_HEIGHT;
								while (showW>4 && showH>4)
								{
									showW-=4;
									showH-=4;
								}
								gameMode=8;
							}
							break;
						default:
							break;
					}
					if (objs[i].type!=0)
						vSpriteSet(sprCount++,(SPRITE*)spr[objs[i].look],objs[i].x-scX,
							objs[i].y-scY);
				}
				for (i=0;i<lives;i++) // Number of lives left
					vSpriteSet(sprCount++,(SPRITE*)spr[27],SCREEN_WIDTH-10*i-12,0);
		
				vMapSetXY(scX,scY);
				vUpdateSpriteMap();

				vitoa(score,tmp,6,'0');
				vPrint(MODE_BLOCK,15,0,tmp); 
				vitoa(level,tmp+1,2,'0');
				tmp[0]='L';
				vPrint(MODE_BLOCK,50,0,tmp); 
				vFlipScreen(1);
				break;
			case 6: // Prepare GAME OVER
				vPrint(MODE_BLOCK,(SCREEN_WIDTH-36)/2,SCREEN_HEIGHT/2,gameOver);
				vFlipScreen(1);
				counter=0;
				gameMode=7;
				break;
			case 7: // GAME OVER pause
				vFlipScreen(1);
				if (counter>80)
				{
					if (score>highscores[9])
					{
						highscores[9]=score;
						i=9;
						while (i>0 && score>highscores[i-1])
						{
							highscores[i]=highscores[i-1];
							i--;
							highscores[i]=score;
						}
						vClearScreen(vRGB(0,0,0));
						for (j=0;j<10;j++)
						{
							if (j==i)
								vSetForeColor(vRGB(255,255,0));
							else
								vSetForeColor(vRGB(255,255,255));
							vitoa((j+1),tmp,2,' ');
							tmp[2]='.';
							tmp[3]='\0';
							vPrint(MODE_BLOCK,30,10+7*j,tmp);
							vitoa(highscores[j],tmp,6,'0');
							vPrint(MODE_BLOCK,SCREEN_WIDTH-54,10+7*j,tmp);
						}
						vSetForeColor(vRGB(255,255,255));
						vPrint(MODE_BLOCK,(SCREEN_WIDTH-40)/2,1,highscore);
						vSpriteClear();
						vSpriteSet(0,(SPRITE*)spr[12],6,7*i+3);
						vUpdateSprite();
						counter=0;
						gameMode=9;
						storeHighscores();
					}
					else
						gameMode=1;
				}
				break;
			case 8: // Fade out, prepare for new level
				vSetForeColor(vRGB(0,0,0));
				vFillRect((SCREEN_WIDTH-showW)/2,(SCREEN_HEIGHT-showH)/2,
					(SCREEN_WIDTH-showW)/2+showW,(SCREEN_HEIGHT-showH)/2+showH);
				vFlipScreen(1);
				showW+=4;
				showH+=4;
				if (showW>SCREEN_WIDTH)
				{
					vSetForeColor(vRGB(255,255,255));
					level++;
					gameMode=3;
				}
				break;
			case 9: // Waiting to return to menu
				vFlipScreen(1);
				if (counter>120)
					gameMode=1;
				break;
			case 10: // Warning screen
				vFlipScreen(1);
				if (counter>45)
				{
					gameMode=4;
					vClearScreen(vRGB(0,0,0));
					vSpriteClear();
				}
				break;
			default:
				break;
		}
	}
}

int main(void) 
{
	initResources();
	readHighscores();
	mainLoop();

	return 0;
}
